package gregapi.tileentity;

import static gregapi.data.CS.F;
import static gregapi.data.CS.NF;
import static gregapi.data.CS.T;
import gregapi.util.UT;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraftforge.fluids.Fluid;
import net.minecraftforge.fluids.FluidStack;
import net.minecraftforge.fluids.FluidTankInfo;
import net.minecraftforge.fluids.IFluidTank;

public class FluidTankGT implements IFluidTank {
	private FluidStack mFluid;
	private int mCapacity;
	private boolean mPreventDraining = F;
	
	public FluidTankGT(FluidStack aFluid, long aCapacity) {mFluid = aFluid; mCapacity = UT.Code.bindInt(aCapacity);}
	public FluidTankGT(long aCapacity) {this(NF, aCapacity);}
	public FluidTankGT(Fluid aFluid, int aAmount, long aCapacity) {this(UT.Fluids.make(aFluid, aAmount), aCapacity);}
	public FluidTankGT(NBTTagCompound aNBT, long aCapacity) {this(UT.Fluids.load(aNBT), aCapacity);}
	
	public FluidTankGT readFromNBT(NBTTagCompound aNBT, String aKey) {if (aNBT.hasKey(aKey)) setFluid(UT.Fluids.load(aNBT.getCompoundTag(aKey))); return this;}
	public NBTTagCompound writeToNBT(NBTTagCompound aNBT, String aKey) {if (mFluid != null && (mPreventDraining || mFluid.amount > 0)) aNBT.setTag(aKey, mFluid.writeToNBT(new NBTTagCompound())); else aNBT.removeTag(aKey); return aNBT;}
	
	@Override
	public int fill(FluidStack aFluid, boolean aDoFill) {
		if (aFluid == null) return 0;
		if (!aDoFill) {
			if (mFluid == null) return Math.min(mCapacity, aFluid.amount);
			if (!mFluid.isFluidEqual(aFluid)) return 0;
			return Math.min(mCapacity - mFluid.amount, aFluid.amount);
		}
		if (mFluid == null) {
			mFluid = UT.Fluids.make(aFluid, Math.min(mCapacity, aFluid.amount));
			return mFluid.amount;
		}
		if (!mFluid.isFluidEqual(aFluid)) return 0;
		int filled = mCapacity - mFluid.amount;
		if (aFluid.amount < filled) {
			mFluid.amount += aFluid.amount;
			filled = aFluid.amount;
		} else mFluid.amount = mCapacity;
		return filled;
	}
	
	@Override
	public FluidStack drain(int aDrained, boolean aDoDrain) {
		if (mFluid == null) return null;
		int drained = aDrained;
		if (mFluid.amount < drained) drained = mFluid.amount;
		FluidStack stack = new FluidStack(mFluid, drained);
		if (aDoDrain) {
			mFluid.amount -= drained;
			if (mFluid.amount <= 0 && !mPreventDraining) mFluid = null;
		}
		return stack;
	}
	
	public FluidTankGT setFluid(FluidStack aFluid) {mFluid = aFluid; return this;}
	public FluidTankGT setCapacity(long aCapacity) {mCapacity = UT.Code.bindInt(aCapacity); return this;}
	public FluidTankGT setPreventDraining() {return setPreventDraining(T);}
	public FluidTankGT setPreventDraining(boolean aPrevent) {mPreventDraining = aPrevent; return this;}
	
	@Override public FluidStack getFluid() {return mFluid;}
	@Override public int getFluidAmount() {return mFluid == null ? 0 : mFluid.amount;}
	@Override public int getCapacity() {return mCapacity;}
	@Override public FluidTankInfo getInfo() {return new FluidTankInfo(mFluid == null ? null : mFluid.copy(), mCapacity);}
}